from util import format_number
from vector2 import Vector2
from vector3 import Vector3

from math import sin, cos, tan, sqrt, pi, radians
from copy import deepcopy

#import psyco
#psyco.profile()

class MatrixError(Exception):
    
    """Matrix Exception class"""
    
    def __init__(self, description):
        self.description = description
        
    def __str__(self):        
        return self.description
        


class Matrix(object):

    _identity = ( (1.,0.,0.,0.), (0.,1.,0.,0.), (0.,0.,1.,0.), (0.,0.,0.,1.) )
    
    
    def __init__(self, *args):
        """If no parameteres are given, the matrix is initialised to the identity matrix.
        If 1 parameter is given it should be an iterable with the 16 values of the matrix.
        If 4 parameters are given they should be 4 iterables each returning up to 4 values.
        Missing values in each row are padded out with values from the identity matix
        (so you can use Vector3's or tuples of 3 values).
        
        """
        
        def take_4(a, b):   
    
            ia, ib = iter(a), iter(b)
            i = 0
            try:
                while i < 4:
                    yield ia.next()
                    ib.next()
                    i+= 1
            except StopIteration:                
                while i < 4:
                    yield ib.next()
                    i+= 1
                    
        if not args:
            self._m = [1.,0.,0.,0., 0.,1.,0.,0., 0.,0.,1.,0., 0.,0.,0.,1.]
            return
                    
            
        elif len(args) == 4:            
            it_values = iter(args)            
            self._m = sum( list( take_4(it_values.next(), Matrix._identity[row_no]) ) for row_no in xrange(4) )
                    
        else:
            raise TypeError("Matrix.__init__() takes 0, or 4 arguments (%i given)"%len(args))
        
         
    def _row_getter(row_no):        
        def get(self):
            return self.get_row(row_no)
        return get
    
    def _row_setter(row_no):
        def set(self, value):
            if len(value) > 4:
                raise TypeError("Column has at most 4 values")
            values = list(value)+list(self.get_row(row_no)[len(value):])
            return self.set_row(row_no, values)
        return set
        
    row0 = property(_row_getter(0), _row_setter(0), None, "Row 0")
    row1 = property(_row_getter(1), _row_setter(1), None, "Row 1")
    row2 = property(_row_getter(2), _row_setter(2), None, "Row 2")
    row3 = property(_row_getter(3), _row_setter(3), None, "Row 3")
    
    x_axis = property(_row_getter(0), _row_setter(0), None, "Row 0")
    right = property(_row_getter(0), _row_setter(0), None, "Row 0")
    y_axis = property(_row_getter(1), _row_setter(1), None, "Row 1")
    up = property(_row_getter(1), _row_setter(1), None, "Row 1")
    z_axis = property(_row_getter(2), _row_setter(2), None, "Row 2")
    heading = property(_row_getter(2), _row_setter(2), None, "Row 2")
    translation = property(_row_getter(3), _row_setter(3), None, "Row 3")
        
       
    def set(self, row1, row2, row3, row4):
                
        self._m[0:4] = row1
        self._m[4:8] = row2
        self._m[8:12] = row3
        self._m[12:16] = row4
        
        
    def get_row(self, row_no):
        
        try:
            return tuple(self._m[row_no*4:row_no*4+4])
        except IndexError:
            raise IndexError("Row must be 0, 1, 2 or 3")
            
        
    def rows(self):
        
        for i in xrange(0, 16, 4):
            yield tuple(self._m[i:i+4])
        
    @staticmethod
    def from_iter(iterable):
        """Creates a Matrix from an iterable of 16 values."""
        
        m = Matrix.__new__(Matrix, object)
        it_values = iter(iterable)
        m._m = map(float, iterable)
        if len(m._m) != 16:
            raise MatrixError("Iterable must have 16 values")
        return m
        
        
    @staticmethod
    def copy(copy_matrix):        
        """Creates a matrix that is a copy of another."""
        
        m = Matrix.__new__(Matrix, object)
        m._m = copy_matrix._m[:]
        return m
            
            
    @staticmethod
    def blank():        
        """Create a blank matrix (with no information). This is rarely required,
        you may want to use an identity matrix, see Matrix.identity()
        
        """
        
        m = Matrix.__new__(Matrix, object)
        m._m = [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,]
        return m
           
            
    @staticmethod
    def identity():        
        """Creates and identity matrix."""
        
        m = Matrix.__new__(Matrix, object)
        m._m = [1., 0., 0., 0., 0., 1., 0., 0., 0., 0., 1., 0., 0., 0., 0., 1.]
        return m
     
     
    @staticmethod     
    def scale(scale_x, scale_y= None, scale_z= None):        
        """Creates a scale matrix.
        If one parameter is given the scale is uniform,
        if three parameters are give the scale is different (potentialy) on each x axis.
        
        """
                
        m = Matrix.__new__(Matrix, object)
        return m.make_scale(scale_x, scale_y, scale_z)
    
    
    @staticmethod
    def translate(x, y, z):
        """Creates a translation matrix to (x, y, z)."""
        
        m = Matrix.__new__(Matrix, object)
        return m.make_translation(x, y, z)
    
    
    @staticmethod
    def x_rotation(angle):
        """Creates a matrix that does a rotation about the x axis."""
            
        m = Matrix.__new__(Matrix, object)
        return m.make_x_rotation(angle)


    @staticmethod
    def y_rotation(angle):
        """Creates a matrix that does a rotation about the y axis."""
        
        m = Matrix.__new__(Matrix, object)
        return m.make_y_rotation(angle)
        
    
    @staticmethod
    def z_rotation(angle):
        """Creates a matrix that does a rotation about the z axis."""
        
        m = Matrix.__new__(Matrix, object)
        return m.make_z_rotation(angle)
    
    
    @staticmethod
    def rotation_about_axis(axis, angle):
        """Creates a matrix that does a rotation about an axis.
        
        Axis can be a vector or any iterable with 3 values.
        
        """
        
        m = Matrix.__new__(Matrix, object)
        return m.make_rotation_about_axis(axis, angle)
        
        
    @staticmethod
    def xyz_rotation(angle_x, angle_y, angle_z):
        """Creates a matrix that does a rotation about each axis."""
        
        m = Matrix.__new__(Matrix, object)
        return m.make_xyz_rotation(angle_x, angle_y, angle_z)
        
        
    @staticmethod
    def perspective_projection(left, right, top, bottom, near, far):
        """Creates a matrix that projects points in to 2d space.
        
        left -- Coordinate of left of screen
        right -- Coordination of right of screen
        top -- Coordination of the top of the screen
        bottom -- Coordination of the borrom of the screen
        near -- Coordination of the near clipping plane
        far -- Coordinate of the far clipping plane
        
        """
        
        m = Matrix.__new__(Matrix, object)
        return m.make_persepctive_projection(left, right, top, bottom, near, far)
    

    @staticmethod
    def perspective_projection_fov(fov, aspect, near, far):
        """Creates a matrix that projects points in to 2d space
        
        fov -- The field of view (in radians)
        aspect -- The aspect ratio of the screen (width / height)
        near -- Coordinate of the near clipping plane
        far -- Coordinate of the far clipping plane
        
        """
        
        m = Matrix.__new__(Matrix, object)
        return m.make_perspective_projection_fov(fov, aspect, near, far)
        
        
    def __str__(self):
        """'Pretty' formatting of the matrix."""
    
        max_len = max( len(format_number(v)) for v in self.components() )
        
        def format_row(row):            
            return "%s" % " ".join( format_number(value).ljust(max_len) for value in row )
        
        rows = [ format_row(row).rstrip() for row in self.rows() ]
        max_row_len = max(len(row) for row in rows)
        return "\n".join( "[ %s ]" % row.ljust(max_row_len) for row in rows )
            
            
    def __repr__(self):
                
        def format_row(row):            
            return "(%s)" % ", ".join( format_number(value) for value in row )
                
        return "Matrix(%s)" % ", ".join(format_row(row) for row in self.rows())
        
    
    def __setitem__(self, coord, value):
        """Sets an individual element in the matrix.
        coord is a tuple of (row, column)
        
        eg. matrix[2,3] = 3.
        
        """
        
        try:
            self._m[coord[0] * 4 + coord[1]] = float(value)
        except IndexError:
            raise IndexError( "Row and Column should be 0, 1, 2 or 3!" )
            
        
    def __getitem__(self, coord):
        """Gets an individual element in the matrix.
        coord is a tuple of (row, column)
        
        eg. print matrix[2,3]
        
        """
        
        try:
            return self._m[coord[0] * 4 + coord[1]]            
        except IndexError:
            raise IndexError( "Row and Column should be 0, 1, 2 or 3!" )
        except TypeError:
            raise TypeError( "index should be two values containing the row and column" )
            
            
    def __iter__(self):
        """Iterates over all 16 values in the matrix."""
        
        return iter(self._m)
            
            
    def __mul__(self, rhs):
        """Returns the result of multiplying this matrix by another, called by the * (multiply) operator."""
        
                
        m1 = self._m
        m2 = rhs._m
        
        
        retm = [ m2[0] * m1[0] + m2[1] * m1[4] + m2[2] * m1[8] + m2[3] * m1[12],
               m2[0] * m1[1] + m2[1] * m1[5] + m2[2] * m1[9] + m2[3] * m1[13],
               m2[0] * m1[2] + m2[1] * m1[6] + m2[2] * m1[10] + m2[3] * m1[14],
               m2[0] * m1[3] + m2[1] * m1[7] + m2[2] * m1[11] + m2[3] * m1[15],
               
               m2[4] * m1[0] + m2[5] * m1[4] + m2[6] * m1[8] + m2[7] * m1[12],
               m2[4] * m1[1] + m2[5] * m1[5] + m2[6] * m1[9] + m2[7] * m1[13],
               m2[4] * m1[2] + m2[5] * m1[6] + m2[6] * m1[10] + m2[7] * m1[14],
               m2[4] * m1[3] + m2[5] * m1[7] + m2[6] * m1[11] + m2[7] * m1[15],
               
               m2[8] * m1[0] + m2[9] * m1[4] + m2[10] * m1[8] + m2[11] * m1[12],
               m2[8] * m1[1] + m2[9] * m1[5] + m2[10] * m1[9] + m2[11] * m1[13],
               m2[8] * m1[2] + m2[9] * m1[6] + m2[10] * m1[10] + m2[11] * m1[14],
               m2[8] * m1[3] + m2[9] * m1[7] + m2[10] * m1[11] + m2[11] * m1[15],
               
               m2[12] * m1[0] + m2[13] * m1[4] + m2[14] * m1[8] + m2[15] * m1[12],
               m2[12] * m1[1] + m2[13] * m1[5] + m2[14] * m1[9] + m2[15] * m1[13],
               m2[12] * m1[2] + m2[13] * m1[6] + m2[14] * m1[10] + m2[15] * m1[14],
               m2[12] * m1[3] + m2[13] * m1[7] + m2[14] * m1[11] + m2[15] * m1[15] ]

        ret = Matrix.__new__(Matrix, object)
        ret._m = retm

        return ret
        
        
    def __imul__(self, rhs):
        """Multiplies this matrix by another, called by the *= operator."""
                
        m1 = self._m 
        m2 = rhs._m
        
        self._m =    [ m2[0] * m1[0] + m2[1] * m1[4] + m2[2] * m1[8] + m2[3] * m1[12],
                       m2[0] * m1[1] + m2[1] * m1[5] + m2[2] * m1[9] + m2[3] * m1[13],
                       m2[0] * m1[2] + m2[1] * m1[6] + m2[2] * m1[10] + m2[3] * m1[14],
                       m2[0] * m1[3] + m2[1] * m1[7] + m2[2] * m1[11] + m2[3] * m1[15],
                       
                       m2[4] * m1[0] + m2[5] * m1[4] + m2[6] * m1[8] + m2[7] * m1[12],
                       m2[4] * m1[1] + m2[5] * m1[5] + m2[6] * m1[9] + m2[7] * m1[13],
                       m2[4] * m1[2] + m2[5] * m1[6] + m2[6] * m1[10] + m2[7] * m1[14],
                       m2[4] * m1[3] + m2[5] * m1[7] + m2[6] * m1[11] + m2[7] * m1[15],
                       
                       m2[8] * m1[0] + m2[9] * m1[4] + m2[10] * m1[8] + m2[11] * m1[12],
                       m2[8] * m1[1] + m2[9] * m1[5] + m2[10] * m1[9] + m2[11] * m1[13],
                       m2[8] * m1[2] + m2[9] * m1[6] + m2[10] * m1[10] + m2[11] * m1[14],
                       m2[8] * m1[3] + m2[9] * m1[7] + m2[10] * m1[11] + m2[11] * m1[15],
                       
                       m2[12] * m1[0] + m2[13] * m1[4] + m2[14] * m1[8] + m2[15] * m1[12],
                       m2[12] * m1[1] + m2[13] * m1[5] + m2[14] * m1[9] + m2[15] * m1[13],
                       m2[12] * m1[2] + m2[13] * m1[6] + m2[14] * m1[10] + m2[15] * m1[14],
                       m2[12] * m1[3] + m2[13] * m1[7] + m2[14] * m1[11] + m2[15] * m1[15] ]
        
        return self
    
    
    def __copy__(self):
        
        return Matrix.Copy(self)
    
    
    def clone(self):
        
        return Matrix.Copy(self)


    def components(self):
        """Returns an iterator for the components in the matrix. ie returns all 16 values."""
                
        return iter(self._m)

                
    def transposed_components(self):
        """Returns an iterator for the components in the matrix in transposed order."""
        
        for col in self.columns():
            for value in col:
                yield value
                
                
    def rows(self):
        """Returns an iterator for the rows in the matrix (yields 4 tuples of 4 values)."""
        
        for i in xrange(0, 16, 4):
            yield tuple(self._m[i:i+4])
            
            
    def columns(self):
        """Returns an iterator for the columns in the matrix (yields 4 tuples of 4 values)."""
        
        for col in xrange(4):
            yield tuple( self._m[row*4+col] for row in xrange(4) )
    
    
    def row(self, row_no):
        """Returns a row as a tuple of 4 values.
        
        row_no -- The row index
        
        """
        
        try:
            return tuple(self._m[row_no*4:row_no*4+4])
        except IndexError:
            raise IndexError( "Row and Column should be 0, 1, 2 or 3" )
        
        
    def row_vec(self, row_no):
        """Returns a Vector3 for a given row.
        
        row_no -- The row index
        
        """        
        
        try:
            return Vector3(self._m[row_no*4:row_no*4+3])
        except IndexError:
            raise IndexError( "Row and Column should be 0, 1, 2 or 3" )
        
            
    def column(self, col_no):
        """Returns a column as a tuple of 4 values.
        
        col_no -- The column index
        
        """
        
        try:
            return tuple( self._m[row_no*4+colno] for row_no in xrange(4) )
        except IndexError:
            raise IndexError( "Column should be 0, 1, 2 or 3" )
    
    
    def set_row(self, row_no, row):
        """Sets the values in a row.
        
        row_no -- The index of the row
        row -- An container containing the new values
        
        """
        
        try:            
            self._m[row_no*4:row_no*4+4] = map(float, row)
        except IndexError:
            raise IndexError( "Row should be 0, 1, 2 or 3" )
        
        
    def set_column(self, col_no, col):
        """Sets the values in a column.
        
        col_no -- The index of the column
        col -- An iterable containing the new values
        
        """
        
        try:
            col_iter = iter(map(float, col))
            for row_no in xrange(len(col)):
                self._m[row_no*4+col_no] = col_iter.next()
        except IndexError:
            raise IndexError( "Column should be 0, 1, 2 or 3" )
    
    
    def transform(self, v):
        """Transforms a Vector3 and returns the result."""
        
        m = self._m
        if isinstance(v, Vector3):
            x = v._x
            y = v._y
            z = v._z
        else:
            x, y, z = v
        return Vector3.from_3_floats( x * m[0] + y * m[4] + z * m[8] + m[12],
                                      x * m[1] + y * m[5] + z * m[9] + m[13],
                                      x * m[2] + y * m[6] + z * m[10] + m[14] )        
        
        #return Vector3(x, y, z)
    
        
    def rotate(self, v):        
        """Rotates a Vector3 and returns the result.
        The translation part of the matrix is ignored.
        """
        
        m = self._m
        if isinstance(v, Vector3):
            x = v._x
            y = v._y
            z = v._z
        else:
            x, y, z = v
        x = x * m[0] + y * m[4] + z * m[8]
        y = x * m[1] + y * m[5] + z * m[9]
        z = x * m[2] + y * m[6] + z * m[10]
        
        return Vector3.from_3_floats(x, y, z)
        

    def inverse_transform(self, v):        
        """Inverse trasforms a Vector3 and returns the result.
        Warning: This is expensive, pre-calculate an inverse matrix if you can.
        """
                
        return self.get_inverse().transform(v)
        
    
    def make_identity(self):
        """Makes an identity matrix."""
        
        self._m = [1., 0., 0., 0., 0., 1., 0., 0. ,0., 0., 1., 0., 0., 0., 0., 1.]
        return self
            
            
    def make_copy(self, other):
        """Makes a copy of another matrix."""
        
        self._m = other._m[:]
        
            
    def make_scale(self, scale_x, scale_y= None, scale_z= None):
        """Makes a scale matrix.
        
        If the scale_y and scale_z parameters are not given they default to the same as scale_x.
        
        """
        
        scale_y = scale_y or scale_x
        scale_z = scale_z or scale_x        
        self._m =   [float(scale_x), 0.,             0.,             0.,
                     0.,             float(scale_y), 0.,             0.,
                     0.,             0.,             float(scale_z), 0.,
                     0.,             0.,             0.,             1.]
        return self
    
    
    def make_translation(self, x, y, z):
        """Makes a translation matrix."""
            
        self._m =   [1.,       0.,       0.,       0.,
                     0.,       1.,       0.,       0.,
                     0.,       0.,       1.,       0.,
                     float(x), float(y), float(z), 1.]
        return self
        
        
    def make_x_rotation(self, angle):
        """Makes a rotation matrix around the x axis."""
        
        cos_a = cos(angle)
        sin_a = sin(angle)
        
        self._m =  [1.,  0.,      0.,     0.,
                    0.,  cos_a,   -sin_a,  0.,
                    0.,  sin_a,  cos_a,  0.,
                    0.,  0.,      0.,     1.]
        return self
    
    def make_y_rotation(self, angle):
        """Makes a rotation matrix around the y axis."""
                
        cos_a = cos(angle)
        sin_a = sin(angle)
       
        self._m =  [cos_a,  0.,  sin_a, 0.,
                    0.,     1.,  0.,     0.,
                    -sin_a,  0.,  cos_a,  0.,
                    0.,     0.,  0.,     1.] 
        return self
    
       
    def make_z_rotation(self, angle):
        """Makes a rotation matrix around the z axis."""
                
        cos_a = cos(angle)
        sin_a = sin(angle)
       
        self._m =  [ cos_a,   -sin_a,  0.,  0.,
                     sin_a,   cos_a,  0.,  0.,
                     0.,      0.,     1.,  0.,
                     0.,      0.,     0.,  1.]
        return self
        
        
    def make_rotation_about_axis(self, axis, angle):
        """Makes a rotation matrix around an axis.
        
        axis -- An iterable containing the axis (three values)
        angle -- The angle to rotate (in radians)
        
        """
                
        c = cos(angle)
        s = sin(angle)        
        omc = 1. - c        
        x, y, z = axis
        
        self._m = [x*x*omc+c,   y*x*omc+z*s, x*z*omc-y*s, 0.,
                   x*y*omc-z*s, y*y*omc+c,   y*z*omc+x*s, 0.,
                   x*z*omc+y*s, y*z*omc-x*s, z*z*omc+c,   0.,
                   0.,          0.,          0.,          1.]
        return self
        
   
    def make_xyz_rotation(self, angle_x, angle_y, angle_z):
        """Makes a rotation matrix about 3 axis."""
        
        cx = cos(-angle_x)
        sx = sin(-angle_x)
        cy = cos(-angle_y)
        sy = sin(-angle_y)
        cz = cos(-angle_z)
        sz = sin(-angle_z)
        
        sxsy = sx*sy
        sxcy = sx*cy
        
#        Cos(z).Cos(y) + Sin(z).Sin(x).Sin(y)	Sin(z).Cos(x)	Cos(z).-Sin(y) + Sin(z).Sin(x).Cos(y)
#-Sin(z).Cos(y) + Cos(z).Sin(x).Sin(y)	Cos(z).Cos(x)	Sin(z).Sin(y) + Cos(z).Sin(x).Cos(y)
#Cos(x).Sin(y)	-Sin(x)	Cos(x).Cos(y)
                
        self._m = [cz*cy + sz*sxsy,   sz*cx,  cz*-sy+sz*sxcy, 0.,
                   -sz*cy + cz*sxsy,  cz*cx,  sz*sy+cz*sxcy,  0.,
                   cx*sy,             -sx,    cx*cy,          0.,
                   0.,                0.,     0.,             1.]
        
        
        #self.set   (   [cz*cy + sz*sxsy,  cy*-sz + cz*sxsy, cx*sy, 0.],
        #                [sz*cx,            cz*cx,            -sx,   0.],
        #                [-sz*cz + sz*sxcy, sz*sy + cz*sxcy,  cx*cy, 0.],
        #                [0.,               0.,               0.,    1.] )
        
        return self
    

    def make_perspective_projection(left, right, top, bottom, near, far):
        """Makes a perspective projection matrix.
        
        left -- Coordinate of left of screen
        right -- Coordination of right of screen
        top -- Coordination of the top of the screen
        bottom -- Coordination of the borrom of the screen
        near -- Coordination of the near clipping plane
        far -- Coordinate of the far clipping plane

        """       
                
        self._m = [(2.*near)/(right-left),    0.,                        0.,                          0.,
                   0.,                        (2.*near)/(top-bottom),    0.,                          0.,
                   (right+left)/(right-left), (top+bottom)/(top-bottom), -((far+near)/(far-near)),   -1.,
                   0.,                        0.,                        -((2.*far*near)/(far-near)), 0.]
    
    
    def make_perspective_projection_fov(fov, aspect, near, far):
        """Creates a matrix that projects points in to 2d space
        
        fov -- The field of view (in radians)
        aspect -- The aspect ratio of the screen (width / height)
        near -- Coordinate of the near clipping plane
        far -- Coordinate of the far clipping plane
        
        """        
        
        assert fov < pi, "The field of view should be less than pi radians (180 degrees)"
        
        right = tan( fov/2. ) * near
        left = - right
        top = right / aspect
        bottom = -top
        self.make_perspective_projection(left, right, bottom, top, near, far)
        
        
    def transpose(self):
        """Swaps the rows for columns."""
        
        self._m = sum(list(self.columns()))
        
        
    def get_transpose(self):
        """Returns a matrix that is a copy of this, but with rows and columns swapped."""
        
        return Matrix(*self.columns())
     
     
    def get_inverse_rot_trans(self):
        """Returns the inverse of a matrix with only rotation and translation."""
                        
        ret = Matrix.copy(self)
        m=ret._m
        i=self._m
                
        
        m[5] = i[4]
        m[10] = i[8]
        m[4] = i[1]
        m[6] = i[9]
        m[8] = i[2]
        m[9] = i[6]
        
        m[12] = ( m[0] * -i[12] + 
                  m[4] * -i[13] + 
                  m[8] * -i[14] )
        
        m[13] = ( m[1] * -i[12] +
                  m[5] * -i[13] + 
                  m[9] * -i[14] )
        
        m[14] = ( m[2] * -i[12] +
                  m[6] * -i[13] + 
                  m[10] * -i[14] )        
        
        return ret
     
     
    def get_inverse(self):
        
        ret = Matrix.copy(self)
        m = ret._m
        i = self._m        
        
        negpos=[0., 0.]
        temp = i[0] * i[5] * i[10]        
        negpos[temp>0.]+= temp
        
        temp = i[1] * i[6] * i[8]
        negpos[temp>0.]+= temp
        
        temp = i[2] * i[4] * i[9]
        negpos[temp>0.]+= temp
        
        temp = -i[2] * i[5] * i[8]
        negpos[temp>0.]+= temp
        
        temp = -i[1] * i[4] * i[10]
        negpos[temp>0.]+= temp
        
        temp = -i[0] * i[6] * i[9]
        negpos[temp>0.]+= temp
        
        det_1 = negpos[0]+negpos[1]
                
        if (det_1 == 0.) or (abs(det_1 / (negpos[1] - negpos[0])) < (2. * 0.00000000000000001) ):
            raise MatrixError("This matrix can not be inverted")
        
        det_1 = 1. / det_1
        m[0] =   ( i[5] * i[10] -
                   i[6] * i[9] ) * det_1

        m[4] = - ( i[4] * i[10] -
                   i[6] * i[8] ) * det_1

        m[8] =   ( i[4] * i[9] -
                   i[5] * i[8] ) * det_1

        m[1] = - ( i[1] * i[10] -
                   i[2] * i[9] ) * det_1

        m[5] =   ( i[0] * i[10] -
                   i[2] * i[8] ) * det_1

        m[9] = - ( i[0] * i[9] -
                   i[1] * i[8] ) * det_1

        m[2] =   ( i[1] * i[6] -
                   i[2] * i[5] ) * det_1

        m[6] = - ( i[0] * i[6] -
                   i[2] * i[4] ) * det_1

        m[10] =   ( i[0] * i[5] -
                    i[1] * i[4] ) * det_1

        m[12] = - ( i[12] * m[0] +
                    i[13] * m[4] +
                    i[14] * m[8] )
        m[13] = - ( i[12] * m[1] +
                    i[13] * m[4] +
                    i[14] * m[9] )
        m[14] = - ( i[12] * m[2] +
                    i[13] * m[6] +
                    i[14] * m[10] )
        
        m[3] = m[7] = m[11] = 0.
        m[15] = 1.
        
        return ret

     
def test():
    
    m = Matrix.xyz_rotation(radians(45), radians(20), radians(0))
        
    print m
    
    print 
    
    #print (10, 20, 30) + Vector3(1,2,3)
    
    m.translation = Vector3(m.translation) + (10, 20, 30)
    
    print m
    
    print 
    
    v = (1., 2., 3.)
    print v
    vt = m.transform(v)
    print vt
    
    vit = m.get_inverse().transform(vt)
        
    print vit
    
    
    print m.inverse_transform(vt)
    m[1,2]=3.    

    print
    
    print m.x_axis
    print m.translation
    m.translation = (1, 2, 3)
    print m

        
        

if __name__ == "__main__":
    
    test()